#!/bin/sh
#
# math_error_test.sh
#
# Build a C program to test conformance of a math function to
# specification, and optionally run the program in order to
# conduct the test
#
# Copyright 2008, Linux Foundation, written by Michael Kerrisk,
# 	<mtk.manpages@gmail.com>
# Licensed under GPLv2 or later.
#
#
# July 2008

# Supported type IDs
T_DOUBLE="d"
T_FLOAT="f"
T_LONG_DOUBLE="ld"
T_LONG="l"
T_LONG_LONG="ll"
T_INT="i"

DEFAULT_SUBNORMAL_SHIFT=4

# Return a string containing the C type for a type ID

type_for()
{
    case "$1" in
    "$T_DOUBLE")	echo "double";;
    "$T_FLOAT")		echo "float";;
    "$T_LONG_DOUBLE")	echo "long double";;
    "$T_LONG")		echo "long";;
    "$T_LONG_LONG")	echo "long long";;
    "$T_INT")		echo "int";;
    *) 			echo "Bad case in type_for()" 1>&2; exit 2;;
    esac
}

# Return a string containing a suitable printf() format specification
# for a type ID

format_for()
{
    case "$1" in
    "$T_DOUBLE")	echo "%.17e";;
    "$T_FLOAT")		echo "%.10e";;
    "$T_LONG_DOUBLE")	echo "%.36Le";;
    "$T_LONG")		echo "%ld";;
    "$T_LONG_LONG")	echo "%lld";;
    "$T_INT")		echo "%d";;
    *) 			echo "Bad case in format_for()" 1>&2; exit 2;;
    esac
}

# Return a string containing a suitable strtoxxxx() function
# for a type ID

strto_type_func()
{
    case "$1" in
    "$T_DOUBLE")	echo "strtod($2, $3)";;
    "$T_FLOAT")		echo "strtof($2, $3)";;
    "$T_LONG_DOUBLE")	echo "strtold($2, $3)";;
    "$T_LONG")		echo "strtol($2, $3, 0)";;
    "$T_INT")		echo "strtol($2, $3, 0)";;
    "$T_LONG_LONG")	echo "strtoll($2, $3, 0)";;
    *) 			echo "Bad case in strto_type_func()" 1>&2; exit 2;;
    esac
}

strto_for()
{
    if test $1 = $opt_arg_type; then
        echo "$str_conv_func($2)"
        return;
    fi
    strto_type_func "$1" "$2" "$3"
}

usage()
{
    echo "Bad usage: $2"
    echo "Usage: $0 [options] [-e <expected-result>] func-name [exec-args]"
    echo "Options can be:"
    echo "    -B           Don't build/compile program (just run a test)"
    echo "    -f <type>    Function argument type" \
                          "($T_DOUBLE (double), $T_FLOAT (float),"
    echo "                 $T_LONG_DOUBLE (long double))"
    echo "                 Default: $T_DOUBLE"
    echo "    -n <num>     Number of input arguments to function (default: 1)"
    echo "    -r <type>    Function result type"
    echo "                 (defaults to argument type (-f))"
    echo "                 Can be: $T_DOUBLE (double), $T_FLOAT (float)," \
                          "$T_LONG_DOUBLE (long double), $T_LONG (long),"
    echo "                 $T_INT (int), $T_LONG_LONG (long long)"
    echo "    -R <format>  Format (printf()) for printing function result"
    echo "    -x <type>    Type for input argument 1"
    echo "                 (defaults to argument type (-f))"
    echo "    -X <format>  Format (printf(() for printing input argument 1"
    echo "    -y <type>    Type for input argument 2"
    echo "                 (defaults to argument type (-f))"
    echo "    -Y <format>  Format (printf(() for printing input argument 2"
    echo "    -z <type>    Type for input argument 3"
    echo "                 (defaults to argument type (-f))"
    echo "    -Z <format>  Format (printf(() for printing input argument 3"
    echo "    -o <type>    Type for output argument"
    echo "                 (defaults to argument type (-f))"
    echo "    -O <format>  Format (printf(() for printing output argument"
    echo "                 (An output argument is always assumed to be the"
    echo "                  last in a function's argument list)"
    echo "    -s <str>     Suffix for function name (e.g., \"_r\")"
    echo "    -q           Make test execution quiet"
    echo ""
    echo "func-name   is the name of the function to be tested (will"
    echo "            be suffixed according to -f value, and -s suffix)"
    echo ""
    echo "exec-args   If these are supplied then run a test of the program"
    echo "            using these arguments.  Each argument can be one of "
    echo "            the following (case is not signficant):"
    echo "            "
    echo "            <number>         A floating-poimt number"
    echo "            [-]inf           An infinity"
    echo "            nan              A NaN"
    echo "            [-]max           Largest negative/positive "
    echo "                             floating-point value of function"
    echo "                             argument type."
    echo "            [-]min           Smallest negative/positive "
    echo "                             floating-point value of function"
    echo "                             argument type."
    echo "            [-]subnormal[:n] A subnormal number where the most"
    echo "                             significant 1 is n bits from the"
    echo "                             start of the significand"
    echo "                             (n defaults to $DEFAULT_SUBNORMAL_SHIFT)"
    echo "            [-]dbl_min       Minimum double value"
    echo "            [-]dbl_max       Maximum double value" 
    echo "            [-]flt_min       Minimum float value"
    echo "            [-]flt_max       Maximum float value"
    echo "            [-]ldbl_min      Minimum long double value"
    echo "            [-]ldbl_max      Maximum long double value"
    echo ""
    echo "The -e option is used to specify the extected results of"
    echo "executing the test program so that it calls the test function"
    echo "with the given exec-args.  It has the form"
    echo ""
    echo "     -e<errno>,<exception>,<result>"
    echo "         errno     is EDOM, ERANGE, or 0 (no error expected)"
    echo "         exception is FE_INVALID, FE_DIVBYZERO, FE_UNDERFLOW,"
    echo "                   FE_OVERFLOW, or 0 (no exception expected)"
    echo "         result can be:"
    echo "               +0, -0, -inf, +inf, nan, normal, subnormal"
    echo "               or an actual numeric value"
    exit 2
}

build_test_program_source()
{
cat >> $filename << EOF
/* $filename (auto-generated at $(date +"%F %T")) */
#define _XOPEN_SOURCE 600
#define _GNU_SOURCE
#include <assert.h>
#include <stdlib.h>
#include <unistd.h>
#include <stdio.h>
#include <errno.h>
#include <string.h>
#include <fenv.h>
#include <limits.h>
#include <math.h>
#include <float.h>
#include <ctype.h>

#define EXIT_test_setup_failed 2

static int verbose = 1;

/* Convert a string to a $arg_type, but allow a few special cases
   in the form of non-numeric strings */

static $arg_type
$str_conv_func(char *str)
{
    char *p;
    char *s;
    $arg_type retval;

    s = strdup(str);
    if (s == NULL) {
        perror("strdup");
        exit(EXIT_test_setup_failed);
    }

    for (p = s; *p; p++)
        *p = toupper(*p);

    if (strcmp(s, "-INF") == 0)
        retval = -HUGE_VAL;
    else if (strcmp(s, "INF") == 0 || strcmp(s, "+INF") == 0)
        retval = HUGE_VAL;
    else if (strcmp(s, "NAN") == 0)
        retval = nan$ftype_suffix("");

EOF

    if test "X$opt_arg_type" = "X$T_FLOAT"; then
        minval="FLT_MIN"
        maxval="FLT_MAX"
    elif test "X$opt_arg_type" = "X$T_LONG_DOUBLE"; then
        minval="LDBL_MIN"
        maxval="LDBL_MAX"
    else
        minval="DBL_MIN"
        maxval="DBL_MAX"
    fi

cat >> $filename << EOF
    else if (strcmp(s, "MIN") == 0 || strcmp(s, "+MIN") == 0)
        retval = $minval;
    else if (strcmp(s, "-MIN") == 0)
        retval = -$minval;
    else if (strcmp(s, "MAX") == 0 || strcmp(s, "+MAX") == 0)
        retval = $maxval;
    else if (strcmp(s, "-MAX") == 0)
        retval = -$maxval;

    else if (strcmp(s, "DBL_MIN") == 0 || strcmp(s, "+DBL_MIN") == 0)
        retval = DBL_MIN;
    else if (strcmp(s, "-DBL_MIN") == 0)
        retval = -DBL_MIN;
    else if (strcmp(s, "DBL_MAX") == 0 || strcmp(s, "+DBL_MAX") == 0)
        retval = DBL_MAX;
    else if (strcmp(s, "-DBL_MAX") == 0)
        retval = -DBL_MAX;

    else if (strcmp(s, "FLT_MIN") == 0 || strcmp(s, "+FLT_MIN") == 0)
        retval = FLT_MIN;
    else if (strcmp(s, "-FLT_MIN") == 0)
        retval = -FLT_MIN;
    else if (strcmp(s, "FLT_MAX") == 0 || strcmp(s, "+FLT_MAX") == 0)
        retval = FLT_MAX;
    else if (strcmp(s, "-FLT_MAX") == 0)
        retval = -FLT_MAX;

    else if (strcmp(s, "LDBL_MIN") == 0 || strcmp(s, "+LDBL_MIN") == 0)
        retval = LDBL_MIN;
    else if (strcmp(s, "-LDBL_MIN") == 0)
        retval = -LDBL_MIN;
    else if (strcmp(s, "LDBL_MAX") == 0 || strcmp(s, "+LDBL_MAX") == 0)
        retval = LDBL_MAX;
    else if (strcmp(s, "-LDBL_MAX") == 0)
        retval = -LDBL_MAX;

    else if (strncmp(s, "SUBNORMAL", 9) == 0 ||
	     strncmp(s, "+SUBNORMAL", 10) == 0 ||
             strncmp(s, "-SUBNORMAL", 10) == 0) {
        char *h;
        int d, j;

        h = strchr(s, ':');
        if (h == NULL)
            d = $DEFAULT_SUBNORMAL_SHIFT;
        else
            d = atoi(h + 1);
        retval = $minval;
        for (j = 0; j < d; j++)
            retval /= 2.0;
        if (s[0] == '-')
            retval = -retval;
    } else if (strchr("+-0123456789", s[0]) == NULL) {
        fprintf(stderr, "Bad argument: %s\n", s);
        exit(EXIT_test_setup_failed);
    } else
        retval = $(strto_type_func $opt_arg_type s NULL);

    free(s);
    return retval;
}

static void
clearErrors(void)
{
    errno = 0;
    feclearexcept(FE_ALL_EXCEPT);
}

static void
checkErrors(int *r_errno, int *r_except)
{
    *r_errno = errno;

    if (verbose) {
        if (*r_errno == 0) {
            fprintf(stderr, "errno == 0\n");
        } else {
            if (*r_errno == EDOM)
                fprintf(stderr, "errno == EDOM\n");
            else if (*r_errno == ERANGE)
                fprintf(stderr, "errno == ERANGE\n");
            else
                fprintf(stderr, "errno == %d\n", *r_errno);
        }
    }

    *r_except = fetestexcept(FE_INVALID | FE_DIVBYZERO | FE_OVERFLOW |
                         FE_UNDERFLOW | FE_INEXACT);

    if (verbose) {
        printf("fetestexcept() says: ");
        if (*r_except & FE_INVALID)
            printf(" FE_INVALID");
        if (*r_except & FE_DIVBYZERO)
            printf(" FE_DIVBYZERO");
        if (*r_except & FE_OVERFLOW)
            printf(" FE_OVERFLOW");
        if (*r_except & FE_UNDERFLOW)
            printf(" FE_UNDERFLOW");
        if (*r_except & FE_INEXACT)
            printf(" FE_INEXACT");
        printf("\n");
    }
}

static void
usage(char *pname, char *msg)
{
    if (msg != NULL)
        fprintf(stderr, "%s\n", msg);
    fprintf(stderr, "Usage: %s [-q] [-e errno,except,class] <args>\n",
                    pname);
    fprintf(stderr, "-q  quiet\n");
    fprintf(stderr, "-e  expected results, expressed via errno, exception,"
                    "and class of function result\n");
    fprintf(stderr, "    errno can be EDOM, ERANGE, or 0\n");
    fprintf(stderr, "    except can be FE_INVALID, FE_DIVBYZERO, "
                    "FE_OVERFLOW, FE_UNDERFLOW, or 0\n");
    fprintf(stderr, "    class can be +0, -0, +inf, -inf, nan, normal, "
                    "or subnormal\n");
    exit(EXIT_test_setup_failed);
}

#define EPSILON 1e-5     /* Floating-point values whose relative difference
                            is smaller than this will be considered equal
                            when checking test results */

int
main(int argc, char *argv[])
{
    int opt;
    int e_errno, e_except, e_class, e_sign;
    int dont_check_errno = 0, dont_check_except = 0, dont_check_class = 0;
    int e_opt_present;
    int r_errno, r_except;
    int exit_status = EXIT_SUCCESS;
    int value_expected = 0;
    char *perrno, *pexcept, *pclass;

    $(type_for $opt_argx_type) x;
EOF

if test $opt_nargs -ge 2; then
    echo "    $(type_for $opt_argy_type) y;" >> $filename
fi
if test $opt_nargs -ge 3; then
    echo "    $(type_for $opt_argz_type) z;" >> $filename
fi

if test $res2_present != 0; then
    echo "    $(type_for $opt_res2_type) result2;" >> $filename
fi

cat >> $filename << EOF
    $(type_for $opt_res_type) result;
    $(type_for $opt_res_type) e_value;
    int int_value_expected = 0;
    int r_class;
    int r_sign;

    e_opt_present = 0;
    while ((opt = getopt(argc, argv, "qe:")) != -1) {
        //printf("opt=%c\n", opt);
        switch (opt) {
        case 'q':
            verbose = 0;
            break;

        case 'e':
            {
            char *c1, *c2, *p;

            e_opt_present = 1;

            for (p = optarg; *p != '\0'; p++)
                *p = toupper(*p);

            c1 = strchr(optarg, ',');
            if (c1 == NULL)
                usage(argv[0], "bad -e spec");
            *c1 = '\0';

            c2 = strchr(c1 + 1, ',');
            if (c2 == NULL)
                usage(argv[0], "bad -e spec");
            *c2 = '\0';

            perrno = optarg;
            pexcept = c1 + 1;
            pclass = c2 + 1;

            if (strcmp(perrno, "*") == 0)
                dont_check_errno = 1;
            else if (strcmp(perrno, "0") == 0)
                e_errno = 0;
            else if (strcmp(perrno, "EDOM") == 0)
                e_errno = EDOM;
            else if (strcmp(perrno, "ERANGE") == 0)
                e_errno = ERANGE;
            else
                usage(argv[0], "Bad errno in -e spec");

            if (strcmp(pexcept, "*") == 0)
                dont_check_except = 1;
            else if (strcmp(pexcept, "0") == 0)
                e_except = 0;
            else if (strcmp(pexcept, "FE_DIVBYZERO") == 0)
                e_except = FE_DIVBYZERO;
            else if (strcmp(pexcept, "FE_INVALID") == 0)
                e_except = FE_INVALID;
            else if (strcmp(pexcept, "FE_OVERFLOW") == 0)
                e_except = FE_OVERFLOW;
            else if (strcmp(pexcept, "FE_UNDERFLOW") == 0)
                e_except = FE_UNDERFLOW;
            else
                usage(argv[0], "Bad except in -e spec");


            if (strcmp(pclass, "*") == 0)
                dont_check_class = 1;
            else if (strcmp(pclass, "+0") == 0) {
                e_class = FP_ZERO;
                e_sign = +1;
            } else if (strcmp(pclass, "-0") == 0) {
                e_class = FP_ZERO;
                e_sign = -1;
            } else if (strcmp(pclass, "+INF") == 0) {
                e_class = FP_INFINITE;
                e_sign = +1;
            } else if (strcmp(pclass, "-INF") == 0) {
                e_class = FP_INFINITE;
                e_sign = -1;
            } else if (strcmp(pclass, "NORMAL") == 0) {
                e_class = FP_NORMAL;
            } else if (strcmp(pclass, "SUBNORMAL") == 0) {
                e_class = FP_SUBNORMAL;
            } else if (strcmp(pclass, "NAN") == 0) {
                e_class = FP_NAN;
            } else {
                char *ep;
                value_expected = 1;
                int_value_expected = 1;
                if (strcmp(pclass, "FP_ILOGB0") == 0) {
                    e_value = FP_ILOGB0;
                } else if (strcmp(pclass, "FP_ILOGBNAN") == 0) {
                    e_value = FP_ILOGBNAN;
                } else if (strcmp(pclass, "INT_MAX") == 0) {
                    e_value = INT_MAX;
                } else {
                    e_value = $(strto_type_func $opt_res_type pclass "&ep");
                    if (*ep != '\0') {
                        fprintf(stderr, "ep=%s\n", ep);
                        usage(argv[0], "Unrecognized class value");
                    }
                    if (islessgreater(remainder$ftype_suffix(e_value, 1.0), 0.0))
                        int_value_expected = 0;
                }
            }
            }
            break;

        default:
            usage(argv[0], NULL);
            exit(EXIT_test_setup_failed);
        }
    }

    assert(argc == optind - 1 + $(expr $opt_nargs + 1));
    x = $(strto_for $opt_argx_type "argv[optind]" NULL);
EOF

(
if test $opt_nargs -ge 2; then
    echo "    y = " $(strto_for $opt_argy_type "argv[optind+1]" NULL) ";"
fi
if test $opt_nargs -ge 3; then
    echo "    z = " $(strto_for $opt_argz_type "argv[optind+2]" NULL) ";"
fi
) >> $filename

cat >> $filename << EOF

    clearErrors();

EOF

(
echo -n "    result = $func_name(x"
if test $opt_nargs -ge 2; then
    echo -n ", y"
fi
if test $opt_nargs -ge 3; then
    echo -n ", z"
fi
if test $res2_present -ne 0; then
    echo -n ", &result2"
fi
echo ");"
) >> $filename

cat >> $filename << EOF
    checkErrors(&r_errno, &r_except);
EOF

(
echo -n "    printf(\"$func_name($opt_argx_format"
if test $opt_nargs -ge 2; then
    echo -n ",$opt_argy_format"
fi
if test $opt_nargs -ge 3; then
    echo -n ",$opt_argz_format"
fi
if test $res2_present -ne 0; then
    echo -n ",&result2"
fi
echo -n ")=$opt_res_format"
if test $res2_present -ne 0; then
    echo -n ", result2=$opt_res2_format"
fi
echo -n "\n\", "
echo -n "x, "
if test $opt_nargs -ge 2; then
    echo -n "y, "
fi
if test $opt_nargs -ge 3; then
    echo -n "z, "
fi
echo -n "result"
if test $res2_present -ne 0; then
    echo -n ", result2"
fi
echo ");"
) >> $filename

(
echo "    if (e_opt_present) "
echo -n "        printf(\"$func_name(%s"
if test $opt_nargs -ge 2; then
    echo -n ",%s"
fi
if test $opt_nargs -ge 3; then
    echo -n ",%s"
fi
if test $res2_present -ne 0; then
    echo -n ",&result2"
fi
echo -n ")"
echo -n "; \", "
echo -n "argv[optind]"
if test $opt_nargs -ge 2; then
    echo -n ", argv[optind + 1]"
fi
if test $opt_nargs -ge 3; then
    echo -n ", argv[optind + 2]"
fi
echo ");"

) >> $filename

cat >> $filename << EOF

    printf("%s ", (r_errno == 0) ? "0" :
                  (r_errno == EDOM) ? "EDOM" :
                  (r_errno == ERANGE) ? "ERANGE" : "???errno???");

    if ((r_except & (FE_DIVBYZERO | FE_UNDERFLOW |
                        FE_OVERFLOW | FE_INVALID)) == 0) {
	printf("0 ");
    } else {
	char buf[1000];

	buf[0] = '\0';

	if (r_except & FE_DIVBYZERO)
	    strcat(buf, "FE_DIVBYZERO ");
        if (r_except & FE_UNDERFLOW)
	    strcat(buf, "FE_UNDERFLOW ");
	if (r_except & FE_OVERFLOW)
	    strcat(buf, "FE_OVERFLOW ");
	if (r_except & FE_INVALID)
	    strcat(buf, "FE_INVALID ");
	if (strlen(buf) == 0) {
	    printf("?-except?:%d \n", r_except);

	} else {
	    int j;
	    int ecnt;

	    buf[strlen(buf) - 1] = '\0';
	    ecnt = 1;
	    for (j = 0; buf[j] != '\0'; j++) {
		if (buf[j] == ' ') {
		    buf[j] = ',';
		    ecnt++;
		}
	    }
	    printf("%s ", buf);
	}
    }

    r_class = fpclassify(result);
    r_sign = signbit(result) ? -1 : 1;

    if (r_class == FP_INFINITE || r_class == FP_ZERO)
        printf("%s", signbit(result) ? "-" : "+");
    printf("%s", (r_class == FP_ZERO) ? "0" :
                  (r_class == FP_NAN) ? "nan" :
                  (r_class == FP_INFINITE) ? "inf" :
                  (r_class == FP_NORMAL) ? "normal" :
                  (r_class == FP_SUBNORMAL) ? "subnormal" : "???class???");

    if (e_opt_present) {
        int errno_passed, except_passed, class_passed, all_passed;

        printf("; ");

        errno_passed = dont_check_errno || (r_errno == e_errno);
        except_passed = dont_check_except ||
                (e_except == 0 && (r_except & ~FE_INEXACT) == 0) ||
                (((r_except & ~FE_INEXACT) & e_except) == e_except);
        class_passed = dont_check_class ||
                       (value_expected && int_value_expected &&
                        !isnan(result) && result == e_value) ||
                       (value_expected && !isnan(result) &&
                        fabsl((result - e_value) / e_value) < EPSILON) ||
                       (r_class == e_class &&
                        ((r_class == FP_NAN || r_class == FP_NORMAL ||
                          r_class == FP_SUBNORMAL) ||
                         r_sign == e_sign));

        printf("%s ", (r_errno == e_errno) ? "ok" : "x-errno");
        printf("%s ", except_passed ? "ok" : "x-except");
        printf("%s", class_passed ? "ok" : "x-result");

        all_passed = errno_passed && except_passed && class_passed;

        if (!all_passed)
            exit_status = EXIT_FAILURE;

        printf("; %s", all_passed ? "PASS" : "FAIL");
        if (!all_passed) {
            printf(" (expected: ");
            if (e_errno == 0 && e_except == 0) {
                printf("noerror");
            } else if (e_errno == EDOM && e_except == FE_INVALID) {
                printf("domain");
            } else if (e_errno == ERANGE && e_except == FE_OVERFLOW) {
                printf("overflow");
            } else if (e_errno == ERANGE && e_except == FE_UNDERFLOW) {
                printf("underflow");
            } else if (e_errno == ERANGE && e_except == FE_DIVBYZERO) {
                printf("pole");
            }
            printf(" (%s, %s); %s", perrno, pexcept, pclass);

            printf(")");
        }
    }

    printf("\n\n");
    exit(exit_status);
}
EOF

# echo cc -Wall $filename -o $exefilename -lm
cc -Wall $filename -o $exefilename -lm
if test $? -ne 0; then
    exit 2;
fi
}

opt_arg_type="$T_DOUBLE"
opt_nargs=1
opt_suffix=""

opt_res_type_present=0
opt_res_format_present=0
opt_res2_type_present=0
opt_res2_format_present=0
opt_argx_type_present=0
opt_argx_format_present=0
opt_argy_type_present=0
opt_argy_format_present=0
opt_argz_type_present=0
opt_argz_format_present=0

expected_result_option=""
quiet_option=""
no_build=0

while getopts "Be:f:n:qs:o:O:r:R:x:X:y:Y:z:Z:" opt; do
    case "$opt" in
    f) 	opt_arg_type=$OPTARG;
        if test "X$opt_arg_type" != "X$T_DOUBLE" && \
                test "X$opt_arg_type" != "X$T_FLOAT" && \
                test "X$opt_arg_type" != "X$T_LONG_DOUBLE"
            then
                usage $0 "Wrong function type: $OPTARG"
            fi
        ;;
    e) 	expected_result_option="-e $OPTARG"
        ;;
    q) 	quiet_option="-q"
        ;;
    B) 	no_build=1
        ;;
    n) 	opt_nargs=$OPTARG
        if test $opt_nargs -gt 3; then
            usage $0 "Too many arguments"
        fi
        ;;
    s)	opt_suffix=$OPTARG
        ;;
    o)	opt_res2_type=$OPTARG
        opt_res2_type_present=1
        ;;
    O)	opt_res2_format=$OPTARG
        opt_res2_format_present=1
        ;;
    r)	opt_res_type=$OPTARG
        opt_res_type_present=1
        ;;
    R)	opt_res_format=$OPTARG
        opt_res_format_present=1
        ;;
    x)	opt_argx_type=$OPTARG
        opt_argx_type_present=1
        ;;
    X)	opt_argx_format=$OPTARG
        opt_argx_format_present=1
        ;;
    y)	opt_argy_type=$OPTARG
        opt_argy_type_present=1
        ;;
    Y)	opt_argy_format=$OPTARG
        opt_argy_format_present=1
        ;;
    z)	opt_argz_type=$OPTARG
        opt_argz_type_present=1
        ;;
    Z)	opt_argz_format=$OPTARG
        opt_argz_format_present=1
        ;;
    *)  usage
        ;;
    esac
done
shift `expr $OPTIND - 1`

if test $# -lt 1; then
    usage $0 "Missing fname argument"
fi
fname="$1"

if test "X$opt_arg_type" = "X$T_DOUBLE"; then
    ftype_suffix=""
elif test "X$opt_arg_type" = "X$T_FLOAT"; then
    ftype_suffix="f"
elif test "X$opt_arg_type" = "X$T_LONG_DOUBLE"; then
    ftype_suffix="l"
else
    usage $0 "Unrecognized type"
fi

arg_type=$(type_for $opt_arg_type)
arg_format=$(format_for $opt_arg_type)

if test $opt_res_type_present -eq 0; then
    opt_res_type=$opt_arg_type
fi
if test $opt_res_format_present -eq 0; then
    opt_res_format=$(format_for $opt_res_type)
fi

if test $opt_res2_type_present -eq 0; then
    opt_res2_type=$opt_arg_type
fi
if test $opt_res2_format_present -eq 0; then
    opt_res2_format=$(format_for $opt_res2_type)
fi
res2_present=0
if test $opt_res2_type_present -ne 0 || test $opt_res2_format_present -ne 0; then
    res2_present=1
fi

if test $opt_argx_type_present -eq 0; then
    opt_argx_type=$opt_arg_type
fi
if test $opt_argx_format_present -eq 0; then
    opt_argx_format=$(format_for $opt_argx_type)
fi

if test $opt_argy_type_present -eq 0; then
    opt_argy_type=$opt_arg_type
fi
if test $opt_argy_format_present -eq 0; then
    opt_argy_format=$(format_for $opt_argy_type)
fi

if test $opt_argz_type_present -eq 0; then
    opt_argz_type=$opt_arg_type
fi
if test $opt_argz_format_present -eq 0; then
    opt_argz_format=$(format_for $opt_argz_type)
fi

func_name="$fname$ftype_suffix$opt_suffix"
filename="/tmp/mt_$func_name.c"
exefilename="$(echo $filename | sed 's/\.c$//')"

# echo "opt_arg_type=$opt_arg_type; ftype_suffix=$ftype_suffix"

str_conv_func="string_to_$opt_arg_type"

# echo "str_conv_func=$str_conv_func"

if test $no_build -eq 0; then

    rm -f $filename
    echo "Producing $filename; $exefilename"

    build_test_program_source

fi

if test $# -gt 1; then
    shift 1
#    echo $exefilename $quiet_option $expected_result_option -- $*
    $exefilename $quiet_option $expected_result_option -- $*
    exit $?
fi
