#!/bin/sh
#
# math_tests.sh
#
# Copyright 2008, Linux Foundation, written by Michael Kerrisk,
# 	<mtk.manpages@gmail.com>
# Licensed under GPLv2 or later.
#
# $1 can be d / f /ld to run tests for double, float, or double-precision
# versions of math functions.
#
#
PI=3.14159265358979323846264338327950288419716939937510
PI_DIV_2=1.57079632679489661923132169163975144209858469968755
PI_DIV_4=0.78539816339744830961566084581987572104929234984377
PI_3QUARTER=2.35619449019234492884698253745962716314787704953132
#
T_DOUBLE="d"
T_FLOAT="f"
T_LONG_DOUBLE="ld"
#
ftype="$T_DOUBLE"
if test $# -ge 1; then
    ftype=$1
fi

tflag="-f $ftype"

start_test()
{
    tfunc="$1"
    pass_cnt=0
    fail_cnt=0
    echo ""
}

run_test()
{

    echo "run_test $*"
    case "$1" in
    noerror)
	err="0,0"
	;;
    domain)
	err="edom,fe_invalid"
	;;
    pole)
	err="erange,fe_divbyzero"
	;;
    underflow)
	err="erange,fe_underflow"
	;;
    overflow)
	err="erange,fe_overflow"
	;;
    *)	echo "Bad case in start_test()"
	exit 2
        ;;
    esac

    e_opt="-e$err,$2"

    shift 2
    ./math_error_test.sh -B -q $tflag $e_opt "$@"
    tres=$?
    if test $tres -gt 1; then
	echo "Test setup failed"
	exit 1
    fi
    if test $tres -eq 0; then
	pass_cnt=$(expr $pass_cnt + 1)
    else
	fail_cnt=$(expr $fail_cnt + 1)
    fi
}

test_report()
{
    echo -n "SUMMARY: $tfunc passed=$pass_cnt; failed=$fail_cnt; "
    if test $fail_cnt -gt 0; then
	echo "FAILED"
    else
	echo "PASSED"
    fi
    echo ""
}

#

date
uname -m -s -r
getconf GNU_LIBC_VERSION 2>&1
/lib/libc.so.6 | grep version | grep -i lib

echo "==================== cos ===================="
./math_error_test.sh $tflag -- cos
start_test cos
#          expected   expected	    math_error_test.sh	  function test
#           error      result		options                    args
run_test   noerror	nan				-- cos     nan
run_test   domain	nan				-- cos     +inf
run_test   domain	nan				-- cos     -inf
test_report

echo "==================== sin ===================="
./math_error_test.sh $tflag -- sin
start_test sin
run_test   noerror	nan				-- sin     nan
run_test   domain	nan				-- sin     +inf
run_test   domain	nan				-- sin     -inf
test_report

echo "==================== tan ===================="
./math_error_test.sh $tflag -- tan
start_test tan
run_test   noerror	nan				-- tan     nan
run_test   domain	nan				-- tan     +inf
run_test   domain	nan				-- tan     -inf
# No test case possible for overflow, since the best approximation of
# pi/2 in double precision only yields a tan() value of 1.633e16.
test_report

echo "==================== acos ===================="
./math_error_test.sh $tflag -- acos
start_test acos
run_test   noerror	nan				-- acos    nan
run_test   noerror	+0				-- acos    1
run_test   domain	nan				-- acos    +inf
run_test   domain	nan				-- acos    -inf
run_test   domain	nan				-- acos    -2
run_test   domain	nan				-- acos    2
test_report

echo "==================== asin ===================="
./math_error_test.sh $tflag -- asin
start_test asin
run_test   noerror	nan				-- asin    nan
run_test   noerror	+0				-- asin    +0
run_test   noerror	-0				-- asin    -0
run_test   domain	nan				-- asin    -2
run_test   domain	nan				-- asin    2
test_report

echo "==================== atan ===================="
./math_error_test.sh $tflag -- atan
start_test atan
run_test   noerror	nan				-- atan    nan
run_test   noerror	+0				-- atan    +0
run_test   noerror	-0				-- atan    -0
test_report

echo "==================== atan2 ===================="
./math_error_test.sh $tflag -n2 -- atan2
start_test atan2
run_test   noerror	$PI		-n2		-- atan2   +0 -1
run_test   noerror	-$PI		-n2		-- atan2   -0 -1
run_test   noerror	+0		-n2		-- atan2   +0 1
run_test   noerror	-0		-n2		-- atan2   -0 1
run_test   noerror	-$PI_DIV_2	-n2		-- atan2   -1 +0
run_test   noerror	-$PI_DIV_2	-n2		-- atan2   -1 -0
run_test   noerror	$PI_DIV_2	-n2		-- atan2   1 +0
run_test   noerror	nan		-n2		-- atan2   0 nan
run_test   noerror	nan		-n2		-- atan2   nan 0
run_test   noerror	nan		-n2		-- atan2   nan nan
run_test   noerror	0		-n2		-- atan2   +0 +0
run_test   noerror	-0		-n2		-- atan2   -0 +0
run_test   noerror	$PI		-n2		-- atan2   +0 -0
run_test   noerror	-$PI		-n2		-- atan2   -0 -0
run_test   noerror	$PI		-n2		-- atan2   99 -inf
run_test   noerror	-$PI		-n2		-- atan2   -99 -inf
run_test   noerror	0		-n2		-- atan2   99 +inf
run_test   noerror	-0		-n2		-- atan2   -99 +inf
run_test   noerror	$PI_DIV_2	-n2		-- atan2   +inf 99
run_test   noerror	-$PI_DIV_2	-n2		-- atan2   -inf 99
run_test   noerror	$PI_3QUARTER	-n2		-- atan2   +inf -inf
run_test   noerror	-$PI_3QUARTER	-n2		-- atan2   -inf -inf
run_test   noerror	$PI_DIV_4	-n2		-- atan2   +inf +inf
run_test   noerror	-$PI_DIV_4	-n2		-- atan2   -inf +inf
test_report

echo "==================== cosh ===================="
./math_error_test.sh $tflag -- cosh
start_test cosh
run_test   noerror	nan				-- cosh    nan
run_test   noerror	1				-- cosh    +0
run_test   noerror	1				-- cosh    -0
run_test   noerror	+inf				-- cosh    +inf
run_test   noerror	+inf				-- cosh    -inf
run_test   overflow	inf				-- cosh    max
run_test   overflow	inf				-- cosh    -max
test_report

echo "==================== sinh ===================="
./math_error_test.sh $tflag -- sinh
start_test sinh
run_test   noerror	nan				-- sinh    nan
run_test   noerror	+0				-- sinh    +0
run_test   noerror	-0				-- sinh    -0
run_test   noerror	+inf				-- sinh    +inf
run_test   noerror	-inf				-- sinh    -inf
run_test   overflow	inf				-- sinh    max
run_test   overflow	-inf				-- sinh    -max
test_report

echo "==================== tanh ===================="
./math_error_test.sh $tflag -- tanh
start_test tanh
run_test   noerror	nan				-- tanh    nan
run_test   noerror	+0				-- tanh    +0
run_test   noerror	-0				-- tanh    -0
run_test   noerror	+1				-- tanh    +inf
run_test   noerror	-1				-- tanh    -inf
test_report

echo "==================== acosh ===================="
./math_error_test.sh $tflag -- acosh
start_test acosh
run_test   noerror	nan				-- acosh   nan
run_test   noerror	+0				-- acosh   +1
run_test   noerror	+inf				-- acosh   +inf
run_test   domain	nan				-- acosh   0.5
run_test   domain	nan				-- acosh   0.0
run_test   domain	nan				-- acosh   -1
run_test   domain	nan				-- acosh   -inf
test_report

echo "==================== asinh ===================="
./math_error_test.sh $tflag -- asinh
start_test asinh
run_test   noerror	nan				-- asinh   nan
run_test   noerror	+0				-- asinh   +0
run_test   noerror	-0				-- asinh   -0
run_test   noerror	+inf				-- asinh   +inf
run_test   noerror	-inf				-- asinh   -inf
test_report

echo "==================== atanh ===================="
./math_error_test.sh $tflag -- atanh
start_test atanh
run_test   noerror	nan				-- atanh   nan
run_test   noerror	+0				-- atanh   +0
run_test   noerror	-0				-- atanh   -0
# Pole error delivers wrong errno value:
# http://sourceware.org/bugzilla/show_bug.cgi?id=6759
run_test   pole		+inf				-- atanh   +1
run_test   pole		-inf				-- atanh   -1
run_test   domain	nan				-- atanh   -2
run_test   domain	nan				-- atanh   2
test_report

echo "==================== erf ===================="
./math_error_test.sh $tflag -- erf
start_test erf
run_test   noerror	nan				-- erf     nan
run_test   noerror	+0				-- erf     +0
run_test   noerror	-0				-- erf     -0
run_test   noerror	+1				-- erf     +inf
run_test   noerror	-1				-- erf     -inf
run_test   underflow	"*"				-- erf     subnormal
test_report

echo "==================== erfc ===================="
./math_error_test.sh $tflag -- erfc
start_test erfc
run_test   noerror	nan				-- erfc    nan
run_test   noerror	1				-- erfc    +0
run_test   noerror	1				-- erfc    -0
run_test   noerror	+0				-- erfc    +inf
run_test   noerror	+2				-- erfc    -inf
#
if test $ftype = $T_FLOAT; then
    UF_VAL=10		# Underflow value for "float" function
elif test $ftype = $T_LONG_DOUBLE; then
    UF_VAL=106.9	# Underflow value for "long double" function
else
    UF_VAL=27		# Underflow value for "double" function
fi
run_test   underflow	"*"				-- erfc    $UF_VAL
test_report

echo "==================== exp ===================="
./math_error_test.sh $tflag -- exp
start_test exp
run_test   noerror	nan				-- exp     nan
run_test   noerror	+inf				-- exp     +inf
run_test   noerror	+0				-- exp     -inf
run_test   underflow	+0				-- exp     -10000
run_test   overflow	+inf				-- exp     10000
test_report

echo "==================== exp10 ===================="
./math_error_test.sh $tflag -- exp10
start_test exp10
run_test   noerror	nan				-- exp10   nan
run_test   noerror	+inf				-- exp10   +inf
run_test   noerror	+0				-- exp10   -inf
run_test   underflow	+0				-- exp10   -10000
run_test   overflow	+inf				-- exp10   10000
test_report

echo "==================== exp2 ===================="
./math_error_test.sh $tflag -- exp2
start_test exp2
run_test   noerror	nan				-- exp2    nan
run_test   noerror	+inf				-- exp2    +inf
run_test   noerror	+0				-- exp2    -inf
run_test   underflow	+0				-- exp2    -10000
run_test   overflow	+inf				-- exp2    10000
test_report

echo "==================== expm1 ===================="
./math_error_test.sh $tflag -- expm1
start_test expm1
run_test   noerror	nan				-- expm1   nan
run_test   noerror	+0				-- expm1   +0
run_test   noerror	-0				-- expm1   -0
run_test   noerror	+inf				-- expm1   +inf
run_test   overflow	+inf				-- expm1   1e4
# Some large positive values of x cause expm1() to generate a bogus
# FE_INVALID exception in addition to the (correct) FE_OVERFLOW exception.
# see http://sources.redhat.com/bugzilla/show_bug.cgi?id=6814
run_test   overflow	+inf				-- expm1   1e5
run_test   noerror	-1				-- expm1   -1000000
test_report

echo "==================== ldexp ===================="
./math_error_test.sh $tflag -n2 -y i -- ldexp
start_test ldexp
run_test   noerror	nan		-n2		-- ldexp   nan 1
run_test   noerror	+inf		-n2		-- ldexp   +inf 1
run_test   noerror	-inf		-n2		-- ldexp   -inf 1
run_test   underflow	0		-n2		-- ldexp   1 -100000
run_test   overflow	+inf		-n2		-- ldexp   1 100000
test_report

echo "==================== log ===================="
./math_error_test.sh $tflag -- log
start_test log
run_test   noerror	nan				-- log     nan
run_test   noerror	+0				-- log     1
run_test   noerror	+inf				-- log     +inf
run_test   pole		-inf				-- log     0
run_test   domain	nan				-- log     -1
test_report

echo "==================== log10 ===================="
./math_error_test.sh $tflag -- log10
start_test log10
run_test   noerror	nan				-- log10   nan
run_test   noerror	+0				-- log10   1
run_test   noerror	+inf				-- log10   +inf
run_test   pole		-inf				-- log10   0
run_test   domain	nan				-- log10   -1
test_report

echo "==================== log2 ===================="
./math_error_test.sh $tflag -- log2
start_test log2
run_test   noerror	nan				-- log2    nan
run_test   noerror	+0				-- log2    1
run_test   noerror	+inf				-- log2    +inf
run_test   pole		-inf				-- log2    0
run_test   domain	nan				-- log2    -1
test_report

echo "==================== log1p ===================="
./math_error_test.sh $tflag -- log1p
start_test log1p
run_test   noerror	nan				-- log1p   nan
run_test   noerror	+inf				-- log1p   +inf
run_test   pole		-inf				-- log1p   -1
run_test   domain	nan				-- log1p   -2
test_report

echo "==================== logb ===================="
./math_error_test.sh $tflag -- logb
start_test logb
run_test   noerror	nan				-- logb    nan
run_test   pole		-inf				-- logb    0
run_test   noerror	+inf				-- logb    +inf
run_test   noerror	+inf				-- logb    -inf
test_report

echo "==================== ilogb ===================="
./math_error_test.sh $tflag -r i -- ilogb
start_test ilogb
run_test   domain	FP_ILOGB0	-r i		-- ilogb   0
run_test   domain	FP_ILOGBNAN	-r i		-- ilogb   nan
run_test   domain	INT_MAX		-r i		-- ilogb   +inf
run_test   domain	INT_MAX		-r i		-- ilogb   -inf
test_report

echo "==================== sqrt ===================="
./math_error_test.sh $tflag -- sqrt
start_test sqrt
run_test   noerror	nan				-- sqrt    nan
run_test   noerror	+0				-- sqrt    +0
run_test   noerror	-0				-- sqrt    -0
run_test   noerror	+inf				-- sqrt    +inf
run_test   domain	nan				-- sqrt    -1
test_report

echo "==================== cbrt ===================="
./math_error_test.sh $tflag -- cbrt
start_test cbrt
run_test   noerror	nan				-- cbrt    nan
run_test   noerror	+inf				-- cbrt    +inf
run_test   noerror	-inf				-- cbrt    -inf
run_test   noerror	+0				-- cbrt    +0
run_test   noerror	-0				-- cbrt    -0
test_report

echo "==================== pow ===================="
./math_error_test.sh $tflag -n2 -- pow
start_test pow
run_test   domain	nan		-n2		-- pow     -1 1.5
run_test   overflow	+inf		-n2		-- pow     max 2
run_test   noerror	1		-n2		-- pow     1 nan
run_test   noerror	nan		-n2		-- pow     0 nan
run_test   noerror	+0		-n2		-- pow     +0 3
run_test   noerror	-0		-n2		-- pow     -0 3
run_test   noerror	+0		-n2		-- pow     +0 3.5
run_test   noerror	+0		-n2		-- pow     -0 3.5
run_test   noerror	+0		-n2		-- pow     +0 4
run_test   noerror	+0		-n2		-- pow     -0 4
run_test   noerror	+inf		-n2		-- pow     0.5 -inf
run_test   noerror	+inf		-n2		-- pow     -0.5 -inf
run_test   noerror	+0		-n2		-- pow     1.5 -inf
run_test   noerror	+0		-n2		-- pow     -1.5 -inf
run_test   noerror	+0		-n2		-- pow     0.5 +inf
run_test   noerror	+0		-n2		-- pow     -0.5 +inf
run_test   noerror	+inf		-n2		-- pow     1.5 +inf
run_test   noerror	+inf		-n2		-- pow     -1.5 +inf
# If x is negative, and y has a "large" abosulte value then pow() wrongly
# returns a NaN and flags a domain error
# see bug http://sources.redhat.com/bugzilla/show_bug.cgi?id=3866
run_test   overflow	+inf		-n2		-- pow     0.5 -max
run_test   overflow	+inf		-n2		-- pow     -0.5 -max
run_test   underflow	+0		-n2		-- pow     1.5 -max
run_test   underflow	+0		-n2		-- pow     -1.5 -max
run_test   underflow	+0		-n2		-- pow     0.5 +max
run_test   underflow	+0		-n2		-- pow     -0.5 +max
run_test   overflow	+inf		-n2		-- pow     1.5 +max
run_test   overflow	+inf		-n2		-- pow     -1.5 +max

# For the next two, see http://sources.redhat.com/bugzilla/show_bug.cgi?id=2678
# (POSIX.1 doesn't specify the sign of the zero that results in these cases)
run_test   underflow    -0		-n2		-- pow     -min 3
run_test   underflow    -0		-n2		-- pow     -max -3

run_test   noerror	-0		-n2		-- pow     -inf -3
run_test   noerror	+0		-n2		-- pow     -inf -1e100
run_test   noerror	+0		-n2		-- pow     -inf -3.5
run_test   noerror	-inf		-n2		-- pow     -inf 3
run_test   noerror	+inf		-n2		-- pow     -inf 3.5
run_test   noerror	+0		-n2		-- pow     +inf -0.5
run_test   noerror	+inf		-n2		-- pow     +inf +0.5
# Pole error delivers wrong errno value:
# http://sourceware.org/bugzilla/show_bug.cgi?id=6776
run_test   pole		+inf		-n2		-- pow     +0 -3
run_test   pole		-inf		-n2		-- pow     -0 -3
run_test   pole		+inf		-n2		-- pow     +0 -3.5
run_test   pole		+inf		-n2		-- pow     -0 -3.5
test_report

echo "==================== hypot ===================="
./math_error_test.sh $tflag -n2 -- hypot
start_test hypot
run_test   noerror	+inf		-n2		-- hypot   +inf nan
run_test   noerror	+inf		-n2		-- hypot   -inf nan
run_test   noerror	+inf		-n2		-- hypot   nan +inf
run_test   noerror	+inf		-n2		-- hypot   nan -inf
run_test   noerror	nan		-n2		-- hypot   nan 0
run_test   noerror	nan		-n2		-- hypot   0 nan 
run_test   overflow	+inf		-n2		-- hypot   max max
run_test   underflow	subnormal	-n2		-- hypot   subnormal:8 subnormal:8
test_report

echo "==================== fdim ===================="
./math_error_test.sh $tflag -n2 -- fdim
start_test fdim
run_test   noerror	nan		-n2		-- fdim    +inf nan
run_test   noerror	nan		-n2		-- fdim    nan +inf
run_test   overflow	inf		-n2		-- fdim    max -max
test_report

echo "==================== fmax ===================="
./math_error_test.sh $tflag -n2 -- fmax
start_test fmax
run_test   noerror	1		-n2		-- fmax    1 nan
run_test   noerror	1		-n2		-- fmax    nan 1
run_test   noerror	nan		-n2		-- fmax    nan nan
test_report

echo "==================== fmin ===================="
./math_error_test.sh $tflag -n2 -- fmin
start_test fmin
run_test   noerror	1		-n2		-- fmin    1 nan
run_test   noerror	1		-n2		-- fmin    nan 1
run_test   noerror	nan		-n2		-- fmin    nan nan
test_report

echo "==================== fabs ===================="
./math_error_test.sh $tflag -- fabs
start_test fabs
run_test   noerror	nan				-- fabs    nan
run_test   noerror	+0				-- fabs    -0
run_test   noerror	+inf				-- fabs    +inf
run_test   noerror	+inf				-- fabs    -inf
test_report

echo "==================== ceil ===================="
./math_error_test.sh $tflag -- ceil
start_test ceil
run_test   noerror	nan				-- ceil    nan
run_test   noerror	+0				-- ceil    +0
run_test   noerror	-0				-- ceil    -0
run_test   noerror	+inf				-- ceil    +inf
run_test   noerror	-inf				-- ceil    -inf
test_report

echo "==================== floor ===================="
./math_error_test.sh $tflag -- floor
start_test floor
run_test   noerror	nan				-- floor   nan
run_test   noerror	+0				-- floor   +0
run_test   noerror	-0				-- floor   -0
run_test   noerror	+inf				-- floor   +inf
run_test   noerror	-inf				-- floor   -inf
test_report

echo "==================== trunc ===================="
./math_error_test.sh $tflag -- trunc
start_test trunc
run_test   noerror	nan				-- trunc   nan
run_test   noerror	+0				-- trunc   +0
run_test   noerror	-0				-- trunc   -0
run_test   noerror	+inf				-- trunc   +inf
run_test   noerror	-inf				-- trunc   -inf
test_report

echo "==================== rint ===================="
./math_error_test.sh $tflag -- rint
start_test rint
run_test   noerror	nan				-- rint    nan
run_test   noerror	+0				-- rint    +0
run_test   noerror	-0				-- rint    -0
run_test   noerror	+inf				-- rint    +inf
run_test   noerror	-inf				-- rint    -inf
test_report

echo "==================== round ===================="
./math_error_test.sh $tflag -- round
start_test round
run_test   noerror	nan				-- round   nan
run_test   noerror	+0				-- round   +0
run_test   noerror	-0				-- round   -0
run_test   noerror	+inf				-- round   +inf
run_test   noerror	-inf				-- round   -inf
test_report

echo "==================== lround ===================="
./math_error_test.sh $tflag -r l -- lround
start_test lround
run_test   domain	"*"		-r l		-- lround  nan
run_test   domain	"*"		-r l		-- lround  +inf
run_test   domain	"*"		-r l		-- lround  max
test_report

echo "==================== llround ===================="
./math_error_test.sh $tflag -r ll -- llround
start_test llround
run_test   domain	"*"		-r ll		-- llround nan
run_test   domain	"*"		-r ll		-- llround inf
run_test   domain	"*"		-r ll		-- llround max
test_report

echo "==================== lrint ===================="
./math_error_test.sh $tflag -r l -- lrint
start_test lrint
run_test   domain	"*"		-r l		-- lrint   nan
run_test   domain	"*"		-r l		-- lrint   +inf
run_test   domain	"*"		-r l		-- lrint   max
test_report

echo "==================== llrint ===================="
./math_error_test.sh $tflag -r ll -- llrint
start_test llrint
run_test   domain	"*"		-r ll		-- llrint  nan
run_test   domain	"*"		-r ll		-- llrint  +inf
run_test   domain	"*"		-r ll		-- llrint  max
test_report

echo "==================== nearbyint ===================="
./math_error_test.sh $tflag -- nearbyint
start_test nearbyint
run_test   noerror	nan				-- nearbyint nan
run_test   noerror	+0				-- nearbyint +0
run_test   noerror	-0				-- nearbyint -0
run_test   noerror	+inf				-- nearbyint +inf
run_test   noerror	-inf				-- nearbyint -inf
test_report

echo "==================== nextafter ===================="
./math_error_test.sh $tflag -n2 -- nextafter
start_test nextafter
run_test   noerror	nan		-n2		-- nextafter 1 nan
run_test   noerror	nan		-n2		-- nextafter nan 1
run_test   overflow	+inf		-n2		-- nextafter max +inf
run_test   overflow	-inf		-n2		-- nextafter -max -inf
run_test   underflow	subnormal	-n2		-- nextafter min 0
run_test   underflow	subnormal	-n2		-- nextafter -min 0
test_report

echo "==================== nexttoward ===================="
./math_error_test.sh $tflag -n2 -y ld -- nexttoward
start_test nexttoward
run_test   noerror	nan		-n2 -y ld	-- nexttoward 1 nan
run_test   noerror	nan		-n2 -y ld	-- nexttoward nan 1
run_test   overflow	+inf		-n2 -y ld	-- nexttoward max +inf
run_test   overflow	-inf		-n2 -y ld	-- nexttoward -max -inf
run_test   underflow	subnormal	-n2 -y ld	-- nexttoward min 0
run_test   underflow	subnormal	-n2 -y ld	-- nexttoward -min 0
test_report

echo "==================== fma ===================="
./math_error_test.sh $tflag -n3 -- fma
start_test fma
run_test   noerror	nan		-n3		-- fma     1 nan 1
run_test   noerror	nan		-n3		-- fma     nan 1 1
run_test   domain	nan		-n3		-- fma     +inf 1 -inf
run_test   domain	nan		-n3		-- fma     -inf 1 +inf
run_test   domain	nan		-n3		-- fma     +inf 0 1
run_test   domain	nan		-n3		-- fma     -inf 0 1
run_test   domain	nan		-n3		-- fma     0 +inf 1
run_test   domain	nan		-n3		-- fma     0 -inf 1
run_test   noerror	nan		-n3		-- fma     1 -inf nan
run_test   overflow	+inf		-n3		-- fma     max max 0
run_test   overflow	-inf		-n3		-- fma     max -max 0
run_test   underflow	+0		-n3		-- fma     min min 0
run_test   underflow	-0		-n3		-- fma     min -min 0
test_report

echo "==================== fmod ===================="
./math_error_test.sh $tflag -n2 -- fmod
start_test fmod
run_test   noerror	nan		-n2		-- fmod    1 nan
run_test   noerror	nan		-n2		-- fmod    nan 1
run_test   domain	nan		-n2		-- fmod    +inf 1
run_test   domain	nan		-n2		-- fmod    -inf 1
run_test   domain	nan		-n2		-- fmod    1 +0
run_test   domain	nan		-n2		-- fmod    1 -0
test_report

echo "==================== remainder ===================="
./math_error_test.sh $tflag -n2 -- remainder
start_test remainder
run_test   noerror	nan		-n2		-- remainder +inf nan
run_test   noerror	nan		-n2		-- remainder nan +inf
# The call remainder(nan(""), 0) returns a NaN, as expected,
# but wrongly causes a domain error; it should yield a silent NaN.
# see http://sources.redhat.com/bugzilla/show_bug.cgi
run_test   noerror	nan		-n2		-- remainder nan 0
run_test   domain	nan		-n2		-- remainder inf 1
run_test   domain	nan		-n2		-- remainder 1 0
run_test   domain	nan		-n2		-- remainder inf 0
test_report

echo "==================== remquo ===================="
./math_error_test.sh $tflag -n2 -o i -- remquo
start_test remquo
run_test   noerror	nan		-n2 -o i	-- remquo  +inf nan
run_test   noerror	nan		-n2 -o i	-- remquo  nan +inf
run_test   noerror	nan		-n2 -o i	-- remquo  nan 0
run_test   domain	nan		-n2 -o i	-- remquo  +inf 1
run_test   domain	nan		-n2 -o i	-- remquo  1 0
test_report

echo "==================== scalb ===================="
./math_error_test.sh $tflag -n2 -- scalb
start_test scalb
run_test   noerror	nan		-n2		-- scalb   1 nan
run_test   noerror	nan		-n2		-- scalb   nan 1
run_test   noerror	+inf		-n2		-- scalb   +inf 1
run_test   noerror	-inf		-n2		-- scalb   -inf 1
run_test   noerror	+0		-n2		-- scalb   +0 1
run_test   noerror	-0		-n2		-- scalb   -0 1
run_test   domain	nan		-n2		-- scalb   0 +inf
run_test   domain	nan		-n2		-- scalb   +inf -inf
run_test   domain	nan		-n2		-- scalb   -inf -inf
run_test   overflow	+inf		-n2		-- scalb   max 2000
run_test   overflow	-inf		-n2		-- scalb   -max 2000
run_test   underflow	+0		-n2		-- scalb   min -2000
test_report

echo "==================== scalbln ===================="
./math_error_test.sh $tflag -n2 -y l -- scalbln
start_test scalbln
run_test   noerror	nan		-n2		-- scalbln nan 1
run_test   noerror	+inf		-n2		-- scalbln +inf 1
run_test   noerror	-inf		-n2		-- scalbln -inf 1
run_test   noerror	+0		-n2		-- scalbln +0 1
run_test   noerror	-0		-n2		-- scalbln -0 1
run_test   overflow	+inf		-n2		-- scalbln max 2000
run_test   underflow	+0		-n2		-- scalbln min -2000
test_report

echo "==================== scalbn ===================="
./math_error_test.sh $tflag -n2 -y i -- scalbn
start_test scalbn
run_test   noerror	nan		-n2		-- scalbn  nan 1
run_test   noerror	+inf		-n2		-- scalbn  +inf 1
run_test   noerror	-inf		-n2		-- scalbn  -inf 1
run_test   noerror	+0		-n2		-- scalbn  +0 1
run_test   noerror	-0		-n2		-- scalbn  -0 1
run_test   overflow	+inf		-n2		-- scalbn  max 2000
run_test   underflow	+0		-n2		-- scalbn  min -2000
test_report

echo "==================== j0 ===================="
./math_error_test.sh $tflag -- j0
start_test j0
run_test   noerror	nan				-- j0      nan
run_test   underflow	0				-- j0      max
run_test   underflow	0				-- j0      -max
test_report

echo "==================== j1 ===================="
./math_error_test.sh $tflag -- j1
start_test j1
run_test   noerror	nan				-- j1      nan
run_test   underflow	0				-- j1      max
run_test   underflow	0				-- j1      -max
test_report

echo "==================== jn ===================="
./math_error_test.sh $tflag -n2 -x i -- jn
start_test jn
run_test   noerror	nan		-n2 -x i	-- jn      10 nan
run_test   underflow	0		-n2 -x i	-- jn      10 max
run_test   underflow	0		-n2 -x i	-- jn      10 -max
test_report

echo "==================== y0 ===================="
./math_error_test.sh $tflag -- y0
start_test y0
# Pole error delivers wrong errno value:
# http://sourceware.org/bugzilla/show_bug.cgi?id=6807
run_test   pole         -inf				-- y0      0
run_test   noerror	nan				-- y0      nan
run_test   domain	-inf				-- y0      -1
run_test   underflow	0				-- y0      max
# Not possible to overflow with double
test_report

echo "==================== y1 ===================="
./math_error_test.sh $tflag -- y1
start_test y1
run_test   pole         -inf				-- y0      1
run_test   noerror	nan				-- y1      nan
run_test   domain	-inf				-- y1      -1
run_test   underflow	0				-- y1      max
# Not possible to overflow with double
test_report

echo "==================== yn ===================="
./math_error_test.sh $tflag -n2 -x i -- yn
start_test yn
run_test   pole         -inf				-- yn      1 0
run_test   noerror	nan		-n2 -x i	-- yn      10 nan
run_test   domain	-inf		-n2 -x i	-- yn      10 -1
run_test   underflow	0		-n2 -x i	-- yn      10 max
run_test   overflow	-inf		-n2 -x i	-- yn      1000 min
test_report

echo "==================== tgamma ===================="
./math_error_test.sh $tflag -- tgamma
start_test tgamma
run_test   noerror	nan				-- tgamma  nan
run_test   noerror	+inf				-- tgamma  +inf
run_test   domain	nan				-- tgamma  -1
run_test   domain	nan				-- tgamma  -inf
run_test   overflow	+inf				-- tgamma  1e10
run_test   underflow	0				-- tgamma  -1000000.5
run_test   pole		+inf				-- tgamma  +0
run_test   pole		-inf				-- tgamma  -0
test_report

echo "==================== lgamma ===================="
./math_error_test.sh $tflag -- lgamma
start_test lgamma
run_test   noerror	nan				-- lgamma  nan
run_test   noerror	+0				-- lgamma  1
run_test   noerror	+0				-- lgamma  2
run_test   noerror	+inf				-- lgamma  +inf
run_test   noerror	+inf				-- lgamma  -inf
# Pole error delivers wrong errno value:
# http://sourceware.org/bugzilla/show_bug.cgi?id=6777
run_test   pole		+inf				-- lgamma  -1
run_test   overflow	+inf				-- lgamma  max
test_report
